@@ -1,4 +1,4 @@
-from typing import Optional, Dict, List
+from typing import Optional, Dict, List, Tuple
 from collections import defaultdict
 import logging
 
@@ -85,9 +85,58 @@ class PositionMapping:
         return None
 
 
+class InstructionMappingElement:
+
+    __slots__ = ('ins_addr', 'posmap_pos')
+
+    def __init__(self, ins_addr, posmap_pos):
+        self.ins_addr = ins_addr
+        self.posmap_pos = posmap_pos
+
+    def __contains__(self, offset):
+        return self.ins_addr == offset
+
+    def __repr__(self):
+        return "<%d: %d>" % (self.ins_addr, self.posmap_pos)
+
+
+class InstructionMapping:
+
+    __slots__ = ('_insmap', )
+
+    def __init__(self):
+        self._insmap = SortedDict()
+
+    def items(self):
+        return self._insmap.items()
+
+    def add_mapping(self, ins_addr, posmap_pos):
+        if ins_addr in self._insmap:
+            if posmap_pos <= self._insmap[ins_addr].posmap_pos:
+                self._insmap[ins_addr] = InstructionMappingElement(ins_addr, posmap_pos)
+        else:
+            self._insmap[ins_addr] = InstructionMappingElement(ins_addr, posmap_pos)
+
+    def get_nearest_pos(self, ins_addr):
+        try:
+            pre_max = next(self._insmap.irange(maximum=ins_addr, reverse=True))
+            pre_min = next(self._insmap.irange(minimum=ins_addr, reverse=True))
+        except StopIteration:
+            return None
+
+        e1: InstructionMappingElement = self._insmap[pre_max]
+        e2: InstructionMappingElement = self._insmap[pre_min]
+
+        if abs(ins_addr - e1.ins_addr) <= abs(ins_addr - e2.ins_addr):
+            return e1.posmap_pos
+        else:
+            return e2.posmap_pos
+
+
 class CConstruct:
     """
     Represents a program construct in C.
+    Acts as the base class for all other representation constructions.
     """
 
     __slots__ = ()
@@ -95,17 +144,62 @@ class CConstruct:
     def __init__(self):
         pass
 
-    def c_repr(self, indent=0, posmap=None):
+    def c_repr(self, indent=0, posmap=None, stmt_posmap=None, insmap=None):
+        """
+        Creates the C reperesentation of the code and displays it by
+        constructing a large string. This function is called by each program function that needs to be decompiled.
+        The posmap and stmt_posmap act as position maps for the location of each variable and statment to be
+        tracked for later GUI operations. The stmt_posmap also contains expressions that are nested inside of
+        statments.
 
-        def mapper(chunks, posmap):
+        :param indent:  # of indents (int)
+        :param posmap:
+        :return:
+        """
+
+        def mapper(chunks, posmap, stmt_posmap, insmap):
+            # start all positions at beginning of document
             pos = 0
+
+            # track all cvars to be used only once
+            used_cvars = []
+
+            # get each string and object representation of the chunks
             for s, obj in chunks:
                 if obj is not None:
-                    posmap.add_mapping(pos, len(s), obj)
+                    # first, anything that is a compatible statement must be added
+                    if isinstance(obj, CStatement) and hasattr(obj, 'tags') and \
+                            obj.tags is not None and 'ins_addr' in obj.tags:
+                        stmt_posmap.add_mapping(pos, len(s), obj)
+                        insmap.add_mapping(obj.tags['ins_addr'], pos)
+
+                    # if not a statement, a nested expression works
+                    elif isinstance(obj, CExpression) and hasattr(obj, 'tags') and \
+                            obj.tags is not None and 'ins_addr' in obj.tags:
+
+                        # assure that variables are only placed once in the GUI at the defenition of each var
+                        if isinstance(obj, CVariable):
+                            if obj not in used_cvars:
+                                used_cvars.append(obj)
+                                stmt_posmap.add_mapping(pos, len(s), obj)
+                                insmap.add_mapping(obj.tags['ins_addr'], pos)
+                            else:
+                                posmap.add_mapping(pos, len(s), obj)
+                        else:
+                            stmt_posmap.add_mapping(pos, len(s), obj)
+                            insmap.add_mapping(obj.tags['ins_addr'], pos)
+
+                    # if we are not another type, just add to the default posmap
+                    else:
+                        posmap.add_mapping(pos, len(s), obj)
                 pos += len(s)
                 yield s
 
-        return ''.join(mapper(self.c_repr_chunks(indent), posmap))
+        # A special note about this line:
+        # Polymorphism allows that the c_repr_chunks() call will be called
+        # by the CFunction class, which will then call each statement within it and construct
+        # the chunks that get printed in qccode_edit in angr-management.
+        return ''.join(mapper(self.c_repr_chunks(indent), posmap, stmt_posmap, insmap))
 
     def c_repr_chunks(self, indent=0):
         raise NotImplementedError()
@@ -169,6 +263,8 @@ class CFunction(CConstruct):  # pylint:disable=abstract-method
             yield " ", None
             if variable.name:
                 yield variable.name, cvariable
+            elif isinstance(variable, SimTemporaryVariable):
+                yield "tmp_%d" % variable.tmp_id, cvariable
             else:
                 yield str(variable), cvariable
             yield ";\n", None
@@ -236,7 +332,7 @@ class CExpression:
         if hasattr(expr, 'c_repr_chunks'):
             yield from expr.c_repr_chunks()
         else:
-            yield str(expr), None
+            yield str(expr), expr
 
 
 class CStatements(CStatement):
@@ -294,31 +390,32 @@ class CWhileLoop(CLoop):
     Represents a while loop in C.
     """
 
-    __slots__ = ('condition', 'body', )
+    __slots__ = ('condition', 'body', 'tags',)
 
-    def __init__(self, condition, body):
+    def __init__(self, condition, body, tags=None):
 
         super().__init__()
 
         self.condition = condition
         self.body = body
+        self.tags = tags
 
     def c_repr_chunks(self, indent=0):
 
         indent_str = self.indent_str(indent=indent)
 
-        yield indent_str, None
-        yield "while(", None
+        yield indent_str, self
+        yield "while(", self
         if self.condition is None:
-            yield "true", None
+            yield "true", self
         else:
             yield from self.condition.c_repr_chunks()
-        yield ")\n", None
-        yield indent_str, None
-        yield "{\n", None
+        yield ")\n", self
+        yield indent_str, self
+        yield "{\n", self
         yield from self.body.c_repr_chunks(indent=indent + INDENT_DELTA)
-        yield indent_str, None
-        yield "}\n", None
+        yield indent_str, self
+        yield "}\n", self
 
 
 class CDoWhileLoop(CLoop):
@@ -326,31 +423,32 @@ class CDoWhileLoop(CLoop):
     Represents a do-while loop in C.
     """
 
-    __slots__ = ('condition', 'body', )
+    __slots__ = ('condition', 'body', 'tags',)
 
-    def __init__(self, condition, body):
+    def __init__(self, condition, body, tags=None):
 
         super().__init__()
 
         self.condition = condition
         self.body = body
+        self.tags = tags
 
     def c_repr_chunks(self, indent=0):
 
         indent_str = self.indent_str(indent=indent)
 
-        yield indent_str, None
-        yield "do\n", None
-        yield indent_str, None
-        yield "{\n", None
+        yield indent_str, self
+        yield "do\n", self
+        yield indent_str, self
+        yield "{\n", self
         yield from self.body.c_repr_chunks(indent=indent + INDENT_DELTA)
-        yield indent_str, None
-        yield "} while(", None
+        yield indent_str, self
+        yield "} while(", self
         if self.condition is None:
-            yield "true", None
+            yield "true", self
         else:
             yield from self.condition.c_repr_chunks()
-        yield ");\n", None
+        yield ");\n", self
 
 
 class CIfElse(CStatement):
@@ -358,15 +456,16 @@ class CIfElse(CStatement):
     Represents an if-else construct in C.
     """
 
-    __slots__ = ('condition', 'true_node', 'false_node', )
+    __slots__ = ('condition', 'true_node', 'false_node', 'tags')
 
-    def __init__(self, condition, true_node=None, false_node=None):
+    def __init__(self, condition, true_node=None, false_node=None, tags=None):
 
         super().__init__()
 
         self.condition = condition
         self.true_node = true_node
         self.false_node = false_node
+        self.tags = tags
 
         if self.true_node is None and self.false_node is None:
             raise ValueError("'true_node' and 'false_node' cannot be both unspecified.")
@@ -375,26 +474,26 @@ class CIfElse(CStatement):
 
         indent_str = self.indent_str(indent=indent)
 
-        yield indent_str, None
-        yield "if (", None
+        yield indent_str, self
+        yield "if (", self
         yield from self.condition.c_repr_chunks()
-        yield ")\n", None
-        yield indent_str, None
-        yield "{\n", None
+        yield ")\n", self
+        yield indent_str, self
+        yield "{\n", self
         yield from self.true_node.c_repr_chunks(indent=indent + INDENT_DELTA)
-        yield indent_str, None
-        yield "}\n", None
+        yield indent_str, self
+        yield "}\n", self
 
 
         if self.false_node is not None:
 
-            yield indent_str, None
-            yield "else\n", None
-            yield indent_str, None
-            yield "{\n", None
+            yield indent_str, self
+            yield "else\n", self
+            yield indent_str, self
+            yield "{\n", self
             yield from self.false_node.c_repr_chunks(indent=indent + INDENT_DELTA)
-            yield indent_str, None
-            yield "}\n", None
+            yield indent_str, self
+            yield "}\n", self
 
 
 class CIfBreak(CStatement):
@@ -402,28 +501,29 @@ class CIfBreak(CStatement):
     Represents an if-break statement in C.
     """
 
-    __slots__ = ('condition', )
+    __slots__ = ('condition', 'tags', )
 
-    def __init__(self, condition):
+    def __init__(self, condition, tags=None):
 
         super().__init__()
 
         self.condition = condition
+        self.tags = tags
 
     def c_repr_chunks(self, indent=0):
 
         indent_str = self.indent_str(indent=indent)
 
-        yield indent_str, None
-        yield "if (", None
+        yield indent_str, self
+        yield "if (", self
         yield from self.condition.c_repr_chunks()
-        yield ")\n", None
-        yield indent_str, None
-        yield "{\n", None
-        yield self.indent_str(indent=indent + INDENT_DELTA), None
-        yield "break;\n", None
-        yield indent_str, None
-        yield "}\n", None
+        yield ")\n", self
+        yield indent_str, self
+        yield "{\n", self
+        yield self.indent_str(indent=indent + INDENT_DELTA), self
+        yield "break;\n", self
+        yield indent_str, self
+        yield "}\n", self
 
 
 class CBreak(CStatement):
@@ -431,14 +531,18 @@ class CBreak(CStatement):
     Represents a break statement in C.
     """
 
-    __slots__ = ()
+    __slots__ = ('tags', )
+
+    def __init__(self, tags=None):
+        super().__init__()
+        self.tags = tags
 
     def c_repr_chunks(self, indent=0):
 
         indent_str = self.indent_str(indent=indent)
 
-        yield indent_str, None
-        yield "break;\n", None
+        yield indent_str, self
+        yield "break;\n", self
 
 
 class CContinue(CStatement):
@@ -446,14 +550,18 @@ class CContinue(CStatement):
     Represents a continue statement in C.
     """
 
-    __slots__ = ()
+    __slots__ = ('tags', )
+
+    def __init__(self, tags=None):
+        super().__init__()
+        self.tags = tags
 
     def c_repr_chunks(self, indent=0):
 
         indent_str = self.indent_str(indent=indent)
 
-        yield indent_str, None
-        yield "continue;\n", None
+        yield indent_str, self
+        yield "continue;\n", self
 
 
 class CSwitchCase(CStatement):
@@ -461,39 +569,40 @@ class CSwitchCase(CStatement):
     Represents a switch-case statement in C.
     """
 
-    __slots__ = ('switch', 'cases', 'default', )
+    __slots__ = ('switch', 'cases', 'default', 'tags')
 
-    def __init__(self, switch, cases, default):
+    def __init__(self, switch, cases, default, tags=None):
         super().__init__()
 
         self.switch = switch
         self.cases = cases
         self.default = default
+        self.tags = tags
 
     def c_repr_chunks(self, indent=0):
 
         indent_str = self.indent_str(indent=indent)
 
-        yield indent_str, None
-        yield "switch (", None
+        yield indent_str, self
+        yield "switch (", self
         yield from self.switch.c_repr_chunks()
-        yield ")\n", None
-        yield indent_str, None
-        yield "{\n", None
+        yield ")\n", self
+        yield indent_str, self
+        yield "{\n", self
 
         # cases
         for idx, case in self.cases:
-            yield indent_str, None
-            yield "case {}:\n".format(idx), None
+            yield indent_str, self
+            yield "case {}:\n".format(idx), self
             yield from case.c_repr_chunks(indent=indent + INDENT_DELTA)
 
         if self.default is not None:
-            yield indent_str, None
-            yield "default:\n", None
+            yield indent_str, self
+            yield "default:\n", self
             yield from self.default.c_repr_chunks(indent=indent + INDENT_DELTA)
 
-        yield indent_str, None
-        yield "}\n", None
+        yield indent_str, self
+        yield "}\n", self
 
 
 class CAssignment(CStatement):
@@ -501,24 +610,25 @@ class CAssignment(CStatement):
     a = b
     """
 
-    __slots__ = ('lhs', 'rhs', )
+    __slots__ = ('lhs', 'rhs', 'tags', )
 
-    def __init__(self, lhs, rhs):
+    def __init__(self, lhs, rhs, tags=None):
 
         super().__init__()
 
         self.lhs = lhs
         self.rhs = rhs
+        self.tags = tags
 
     def c_repr_chunks(self, indent=0):
 
         indent_str = self.indent_str(indent=indent)
 
-        yield indent_str, None
+        yield indent_str, self
         yield from CExpression._try_c_repr_chunks(self.lhs)
-        yield " = ", None
+        yield " = ", self
         yield from CExpression._try_c_repr_chunks(self.rhs)
-        yield ";\n", None
+        yield ";\n", self
 
 
 class CFunctionCall(CStatement, CExpression):
@@ -553,11 +663,11 @@ class CFunctionCall(CStatement, CExpression):
 
         indent_str = self.indent_str(indent=indent)
 
-        yield indent_str, None
+        yield indent_str, self
 
         if not self.is_expr and self.ret_expr is not None:
             yield from CExpression._try_c_repr_chunks(self.ret_expr)
-            yield " = ", None
+            yield " = ", self
 
         if self.callee_func is not None:
             if self.callee_func.demangled_name:
@@ -568,62 +678,64 @@ class CFunctionCall(CStatement, CExpression):
         else:
             yield from CExpression._try_c_repr_chunks(self.callee_target)
 
-        yield "(", None
+        yield "(", self
 
         for i, arg in enumerate(self.args):
             if i:
-                yield ", ", None
+                yield ", ", self
             yield from CExpression._try_c_repr_chunks(arg)
 
-        yield ")", None
+        yield ")", self
 
         if not self.is_expr:
-            yield ";", None
+            yield ";", self
             if not self.returning:
-                yield " /* do not return */", None
-            yield "\n",  None
+                yield " /* do not return */", self
+            yield "\n",  self
 
 
 class CReturn(CStatement):
 
-    __slots__ = ('retval', )
+    __slots__ = ('retval', 'tags', )
 
-    def __init__(self, retval):
+    def __init__(self, retval, tags=None):
         super().__init__()
 
         self.retval = retval
+        self.tags = tags
 
     def c_repr_chunks(self, indent=0):
 
         indent_str = self.indent_str(indent=indent)
 
         if not self.retval:
-            yield indent_str, None
-            yield "return;\n", None
+            yield indent_str, self
+            yield "return;\n", self
         else:
-            yield indent_str, None
-            yield "return ", None
+            yield indent_str, self
+            yield "return ", self
             yield from self.retval.c_repr_chunks()
-            yield ";\n", None
+            yield ";\n", self
 
 
 class CGoto(CStatement):
 
-    __slots__ = ('target', )
+    __slots__ = ('target', 'tags', )
 
-    def __init__(self, target):
+    def __init__(self, target, tags=None):
         super().__init__()
 
         self.target = target
+        self.tags = tags
 
     def c_repr_chunks(self, indent=0):
 
         indent_str = self.indent_str(indent=indent)
 
-        yield indent_str, None
-        yield "goto ", None
+        yield indent_str, self
+        yield "goto ", self
         yield from self.target.c_repr_chunks()
-        yield ";\n", None
+        yield ";\n", self
 
 
 class CUnsupportedStatement(CStatement):
@@ -649,15 +761,16 @@ class CUnsupportedStatement(CStatement):
 
 class CStructField(CExpression):
 
-    __slots__ = ('struct_type', 'offset', 'field', )
+    __slots__ = ('struct_type', 'offset', 'field', 'tags', )
 
-    def __init__(self, struct_type, offset, field):
+    def __init__(self, struct_type, offset, field, tags=None):
 
         super().__init__()
 
         self.struct_type = struct_type
         self.offset = offset
         self.field = field
+        self.tags = tags
 
     @property
     def type(self):
@@ -685,15 +798,16 @@ class CVariable(CExpression):
     Read value from a variable.
     """
 
-    __slots__ = ('variable', 'offset', 'variable_type', )
+    __slots__ = ('variable', 'offset', 'variable_type', 'tags', )
 
-    def __init__(self, variable, offset=None, variable_type=None):
+    def __init__(self, variable, offset=None, variable_type=None, tags=None):
 
         super().__init__()
 
         self.variable = variable
         self.offset = offset
         self.variable_type = variable_type
+        self.tags = tags
 
     @property
     def type(self):
@@ -702,9 +816,9 @@ class CVariable(CExpression):
     def _get_offset_string_chunks(self, in_hex=False):
         if type(self.offset) is int:
             if in_hex:
-                yield "%#x" % self.offset, None
+                yield "%#x" % self.offset, self
             else:
-                yield "%d" % self.offset, None
+                yield "%d" % self.offset, self
         else:
             yield from self.offset.c_repr_chunks()
 
@@ -713,6 +827,8 @@ class CVariable(CExpression):
             if isinstance(self.variable, SimVariable):
                 if self.variable.name:
                     yield self.variable.name, self
+                elif isinstance(self.variable, SimTemporaryVariable):
+                    yield "tmp_%d" % self.variable.tmp_id, self
                 else:
                     yield str(self.variable), self
             elif isinstance(self.variable, CExpression):
@@ -723,29 +839,29 @@ class CVariable(CExpression):
                             first_field, *_ = self.variable.type.pts_to.fields
                             c_field = CStructField(self.variable.type.pts_to, 0, first_field)
                             yield from self.variable.c_repr_chunks()
-                            yield "->", None
+                            yield "->", self
                             yield from c_field.c_repr_chunks()
                             return
                         elif isinstance(self.variable.type.pts_to, SimTypeArray):
                             # is it pointing to an array? if so, we take the first element
                             yield from self.variable.c_repr_chunks()
-                            yield "[", None
+                            yield "[", self
                             yield "0", 0
-                            yield "]", None
+                            yield "]", self
                             return
 
                 # default output
-                yield "*(", None
+                yield "*(", self
                 yield from self.variable.c_repr_chunks()
-                yield ")", None
+                yield ")", self
             else:
-                yield str(self.variable), None
+                yield str(self.variable), self
         else:  # self.offset is not None
             if isinstance(self.variable, SimVariable):
                 yield self.variable.name, self
-                yield "[", None
+                yield "[", self
                 yield from self._get_offset_string_chunks()
-                yield "]", None
+                yield "]", self
 
             elif isinstance(self.variable, CExpression):
                 if isinstance(self.variable, CVariable) and self.variable.type is not None:
@@ -759,7 +875,7 @@ class CVariable(CExpression):
                                     field = offset_to_field[self.offset]
                                     c_field = CStructField(t, self.offset, field)
                                     yield from self.variable.c_repr_chunks()
-                                    yield "->", None
+                                    yield "->", self
                                     yield from c_field.c_repr_chunks()
                                     return
                                 else:
@@ -771,36 +887,36 @@ class CVariable(CExpression):
                             if isinstance(self.offset, int):
                                 # it's pointing to an array! take the corresponding element
                                 yield from self.variable.c_repr_chunks()
-                                yield "[", None
+                                yield "[", self
                                 yield str(self.offset), self.offset
-                                yield "]", None
+                                yield "]", self
                                 return
 
                         # other cases
                         yield from self.variable.c_repr_chunks()
-                        yield "[", None
+                        yield "[", self
                         yield from CExpression._try_c_repr_chunks(self.offset)
-                        yield "]", None
+                        yield "]", self
                         return
 
                 # default output
-                yield "*(", None
+                yield "*(", self
                 yield from self.variable.c_repr_chunks()
-                yield ":", None
+                yield ":", self
                 yield from self._get_offset_string_chunks()
-                yield ")", None
+                yield ")", self
 
             elif isinstance(self.variable, Expr.Register):
                 yield self.variable.reg_name if hasattr(self.variable, 'reg_name') else str(self.variable), self
-                yield ":", None
+                yield ":", self
                 yield from self._get_offset_string_chunks(in_hex=True)
 
             else:
-                yield "*(", None
+                yield "*(", self
                 yield str(self.variable), self
-                yield ":", None
+                yield ":", self
                 yield from self._get_offset_string_chunks()
-                yield ")", None
+                yield ")", self
 
 
 class CUnaryOp(CExpression):
@@ -808,15 +924,16 @@ class CUnaryOp(CExpression):
     Unary operations.
     """
 
-    __slots__ = ('op', 'operand', 'variable', )
+    __slots__ = ('op', 'operand', 'variable', 'tags', )
 
-    def __init__(self, op, operand, variable):
+    def __init__(self, op, operand, variable, tags=None):
 
         super().__init__()
 
         self.op = op
         self.operand = operand
         self.variable = variable
+        self.tags = tags
 
     @property
     def type(self):
@@ -829,7 +946,7 @@ class CUnaryOp(CExpression):
 
     def c_repr_chunks(self):
         if self.variable is not None:
-            yield "&", None
+            yield "&", self
             yield from self.variable.c_repr_chunks()
             return
 
@@ -842,19 +959,19 @@ class CUnaryOp(CExpression):
         if handler is not None:
             yield from handler()
         else:
-            yield "UnaryOp %s" % (self.op), None
+            yield "UnaryOp %s" % (self.op), self
 
     #
     # Handlers
     #
 
     def _c_repr_chunks_not(self):
-        yield "!(", None
+        yield "!(", self
         yield from CExpression._try_c_repr_chunks(self.operand)
-        yield ")", None
+        yield ")", self
 
     def _c_repr_chunks_reference(self):
-        yield "&", None
+        yield "&", self
         yield from CExpression._try_c_repr_chunks(self.operand)
 
 
@@ -906,7 +1023,7 @@ class CBinaryOp(CExpression):
     def c_repr_chunks(self):
 
         if self.variable is not None:
-            yield "&", None
+            yield "&", self
             yield from self.variable.c_repr_chunks()
             return
 
@@ -940,7 +1057,7 @@ class CBinaryOp(CExpression):
         if handler is not None:
             yield from handler()
         else:
-            yield "BinaryOp %s" % (self.op), None
+            yield "BinaryOp %s" % (self.op), self
 
     #
     # Handlers
@@ -949,18 +1066,18 @@ class CBinaryOp(CExpression):
     def _c_repr_chunks(self, op):
         # lhs
         if isinstance(self.lhs, CBinaryOp) and self.op_precedence > self.lhs.op_precedence:
-            yield "(", None
+            yield "(", self
             yield from self._try_c_repr_chunks(self.lhs)
-            yield ")", None
+            yield ")", self
         else:
             yield from self._try_c_repr_chunks(self.lhs)
         # operator
-        yield op, None
+        yield op, self
         # rhs
         if isinstance(self.rhs, CBinaryOp) and self.op_precedence > self.rhs.op_precedence - (1 if self.op in ['Sub', 'Div'] else 0):
-            yield "(", None
+            yield "(", self
             yield from self._try_c_repr_chunks(self.rhs)
-            yield ")", None
+            yield ")", self
         else:
             yield from self._try_c_repr_chunks(self.rhs)
 
@@ -1024,15 +1141,16 @@ class CBinaryOp(CExpression):
 
 class CTypeCast(CExpression):
 
-    __slots__ = ('src_type', 'dst_type', 'expr', )
+    __slots__ = ('src_type', 'dst_type', 'expr', 'tags', )
 
-    def __init__(self, src_type, dst_type, expr):
+    def __init__(self, src_type, dst_type, expr, tags=None):
 
         super().__init__()
 
         self.src_type = src_type
         self.dst_type = dst_type
         self.expr = expr
+        self.tags = tags
 
     @property
     def type(self):
@@ -1041,7 +1159,7 @@ class CTypeCast(CExpression):
         return self._type
 
     def c_repr_chunks(self):
-        yield "({})".format(self.dst_type), None
+        yield "({})".format(self.dst_type), self
         yield from CExpression._try_c_repr_chunks(self.expr)
 
 
@@ -1110,26 +1228,27 @@ class CRegister(CExpression):
 
 class CITE(CExpression):
 
-    __slots__ = ('cond', 'iftrue', 'iffalse', )
+    __slots__ = ('cond', 'iftrue', 'iffalse', 'tags', )
 
-    def __init__(self, cond, iftrue, iffalse):
+    def __init__(self, cond, iftrue, iffalse, tags=None):
         super().__init__()
         self.cond = cond
         self.iftrue = iftrue
         self.iffalse = iffalse
+        self.tags = tags
 
     @property
     def type(self):
         return SimTypeInt()
 
     def c_repr_chunks(self):
-        yield "(", None
+        yield "(", self
         yield from self.cond.c_repr_chunks()
-        yield "? ", None
+        yield "? ", self
         yield from self.iftrue.c_repr_chunks()
-        yield " : ", None
+        yield " : ", self
         yield from self.iffalse.c_repr_chunks()
-        yield ")", None
+        yield ")", self
 
 
 class CDirtyExpression(CExpression):
@@ -1198,10 +1317,12 @@ class StructuredCodeGenerator(Analysis):
         self.binop_depth_cutoff = binop_depth_cutoff
 
         self._variables_in_use: Optional[Dict] = None
-        self._memo: Optional[Dict[Expr,CExpression]] = None
+        self._memo: Optional[Dict[Tuple[Expr,bool],CExpression]] = None
 
         self.text = None
         self.posmap = None
+        self.stmt_posmap = None
+        self.insmap = None
         self.nodemap = None
         self._indent = indent
 
@@ -1228,7 +1349,10 @@ class StructuredCodeGenerator(Analysis):
         self._variables_in_use = None
 
         self.posmap = PositionMapping()
-        self.text = func.c_repr(indent=self._indent, posmap=self.posmap)
+        self.stmt_posmap = PositionMapping()
+        self.insmap = InstructionMapping()
+
+        self.text = func.c_repr(indent=self._indent, posmap=self.posmap, stmt_posmap=self.stmt_posmap, insmap=self.insmap)
 
         self.nodemap = defaultdict(set)
         for elem, node in self.posmap.items():
@@ -1295,8 +1419,8 @@ class StructuredCodeGenerator(Analysis):
         l.warning("Unsupported address expression %r", addr)
         return expr, None
 
-    def _cvariable(self, variable, offset=None, variable_type=None):
-        cvariable = CVariable(variable, offset=offset, variable_type=variable_type)
+    def _cvariable(self, variable, offset=None, variable_type=None, tags=None):
+        cvariable = CVariable(variable, offset=offset, variable_type=variable_type, tags=tags)
         if isinstance(variable, SimVariable):
             self._variables_in_use[variable] = cvariable
         return cvariable
@@ -1307,8 +1431,8 @@ class StructuredCodeGenerator(Analysis):
 
     def _handle(self, node, is_expr: bool=True):
 
-        if node in self._memo:
-            return self._memo[node]
+        if (node, is_expr) in self._memo:
+            return self._memo[(node, is_expr)]
 
         handler = self._handlers.get(node.__class__, None)
         if handler is not None:
@@ -1317,7 +1441,7 @@ class StructuredCodeGenerator(Analysis):
                 converted = handler(node, is_expr=is_expr)
             else:
                 converted = handler(node)
-            self._memo[node] = converted
+            self._memo[(node, is_expr)] = converted
             return converted
         raise UnsupportedNodeTypeError("Node type %s is not supported yet." % type(node))
 
@@ -1338,36 +1462,43 @@ class StructuredCodeGenerator(Analysis):
         return CStatements(lines) if len(lines) > 1 else lines[0]
 
     def _handle_Loop(self, loop_node):
+        tags = {'ins_addr': loop_node.addr}
 
         if loop_node.sort == 'while':
             return CWhileLoop(None if loop_node.condition is None else self._handle(loop_node.condition),
-                              self._handle(loop_node.sequence_node, is_expr=False)
+                              self._handle(loop_node.sequence_node, is_expr=False),
+                              tags=tags
                               )
         elif loop_node.sort == 'do-while':
             return CDoWhileLoop(self._handle(loop_node.condition),
-                                self._handle(loop_node.sequence_node, is_expr=False)
+                                self._handle(loop_node.sequence_node, is_expr=False),
+                                tags=tags
                                 )
 
         else:
             raise NotImplementedError()
 
     def _handle_Condition(self, condition_node):
+        tags = {'ins_addr': condition_node.addr}
 
         code = CIfElse(self._handle(condition_node.condition),
                        true_node=self._handle(condition_node.true_node, is_expr=False)
                                  if condition_node.true_node else None,
                        false_node=self._handle(condition_node.false_node, is_expr=False)
                                   if condition_node.false_node else None,
+                       tags=tags
                        )
         return code
 
     def _handle_ConditionalBreak(self, node):  # pylint:disable=no-self-use
+        tags = {'ins_addr': node.addr}
 
-        return CIfBreak(self._handle(node.condition))
+        return CIfBreak(self._handle(node.condition), tags=tags)
 
     def _handle_Break(self, node):  # pylint:disable=no-self-use,unused-argument
+        tags = {'ins_addr': node.addr}
 
-        return CBreak()
+        return CBreak(tags=tags)
 
     def _handle_MultiNode(self, node):  # pylint:disable=no-self-use
 
@@ -1389,12 +1520,14 @@ class StructuredCodeGenerator(Analysis):
         switch_expr = self._handle(node.switch_expr)
         cases = [ (idx, self._handle(case, is_expr=False)) for idx, case in node.cases.items() ]
         default = self._handle(node.default_node, is_expr=False) if node.default_node is not None else None
-        switch_case = CSwitchCase(switch_expr, cases, default=default)
+        tags = {'ins_addr': node.addr}
+        switch_case = CSwitchCase(switch_expr, cases, default=default, tags=tags)
         return switch_case
 
     def _handle_Continue(self, node):  # pylint:disable=no-self-use,unused-argument
+        tags = {'ins_addr': node.addr}
 
-        return CContinue()
+        return CContinue(tags=tags)
 
     def _handle_AILBlock(self, node):
         """
@@ -1462,14 +1595,14 @@ class StructuredCodeGenerator(Analysis):
 
         cdata = self._handle(stmt.data)
 
-        return CAssignment(cvariable, cdata)
+        return CAssignment(cvariable, cdata, tags=stmt.tags)
 
     def _handle_Stmt_Assignment(self, stmt):
 
         cdst = self._handle(stmt.dst)
         csrc = self._handle(stmt.src)
 
-        return CAssignment(cdst, csrc)
+        return CAssignment(cdst, csrc, tags=stmt.tags)
 
     def _handle_Stmt_Call(self, stmt, is_expr: bool=False):
 
@@ -1515,7 +1648,7 @@ class StructuredCodeGenerator(Analysis):
                             reference_values[type_] = self._cfg.memory_data[arg.value]
                     new_arg = CConstant(arg, type_, reference_values=reference_values if reference_values else None,
                                         variable=self._handle(arg.variable) if arg.variable is not None else None,
-                                        tags=arg.tags)
+                                        tags=stmt.tags) #XXX: used from stmt for locating it in the stmt_posmap
                 else:
                     new_arg = self._handle(arg)
                 args.append(new_arg)
@@ -1532,23 +1665,27 @@ class StructuredCodeGenerator(Analysis):
                              )
 
     def _handle_Stmt_Jump(self, stmt):
-        return CGoto(self._handle(stmt.target))
+        return CGoto(self._handle(stmt.target), tags=stmt.tags)
 
     def _handle_Stmt_Return(self, stmt: Stmt.Return):
         if not stmt.ret_exprs:
-            return CReturn(None)
+            return CReturn(None, tags=stmt.tags)
         elif len(stmt.ret_exprs) == 1:
             ret_expr = stmt.ret_exprs[0]
             if ret_expr.variable is not None:
-                return CReturn(self._cvariable(ret_expr.variable, offset=ret_expr.variable_offset))
-            return CReturn(self._handle(ret_expr))
+                return CReturn(self._cvariable(ret_expr.variable, offset=ret_expr.variable_offset),
+                               tags=stmt.tags
+                               )
+            return CReturn(self._handle(ret_expr), tags=stmt.tags)
         else:
             # TODO: Multiple return expressions
             l.warning("StructuredCodeGen does not support multiple return expressions yet. Only picking the first one.")
             ret_expr = stmt.ret_exprs[0]
             if ret_expr.variable is not None:
-                return CReturn(self._cvariable(ret_expr.variable, offset=ret_expr.variable_offset))
-            return CReturn(self._handle(ret_expr))
+                return CReturn(self._cvariable(ret_expr.variable, offset=ret_expr.variable_offset),
+                               tags=stmt.tags
+                               )
+            return CReturn(self._handle(ret_expr), tags=stmt.tags)
 
     #
     # AIL expression handlers
@@ -1573,15 +1710,18 @@ class StructuredCodeGenerator(Analysis):
                 offset = None
             return self._cvariable(expr.variable, offset=offset,
                                    variable_type=self._get_variable_type(expr.variable),
+                                   tags=expr.tags
                                    )
 
         variable, offset = self._parse_load_addr(expr.addr)
 
         if variable is not None:
             return self._cvariable(variable, offset=offset,
-                                   variable_type=self._get_variable_type(variable))
+                                   variable_type=self._get_variable_type(variable),
+                                   tags=expr.tags
+                                   )
         else:
-            return self._cvariable(CConstant(offset, SimTypePointer(SimTypeInt)))  # TODO: Correctly setup tags for the Constant
+            return self._cvariable(CConstant(offset, SimTypePointer(SimTypeInt)), tags=expr.tags)
 
     def _handle_Expr_Tmp(self, expr):  # pylint:disable=no-self-use
 
@@ -1598,6 +1738,7 @@ class StructuredCodeGenerator(Analysis):
 
         return CUnaryOp(expr.op, self._handle(expr.operand),
                         variable=self._handle(expr.variable) if expr.variable is not None else None,
+                        tags=expr.tags
                         )
 
     def _handle_Expr_BinaryOp(self, expr):
@@ -1629,21 +1770,21 @@ class StructuredCodeGenerator(Analysis):
         else:
             raise UnsupportedNodeTypeError("Unsupported conversion bits %s." % expr.to_bits)
 
-        return CTypeCast(None, dst_type, self._handle(expr.operand))
+        return CTypeCast(None, dst_type, self._handle(expr.operand), tags=expr.tags)
 
     def _handle_Expr_Dirty(self, expr):  # pylint:disable=no-self-use
         return CDirtyExpression(expr)
 
     def _handle_Expr_ITE(self, expr: Expr.ITE):
-        return CITE(self._handle(expr.cond), self._handle(expr.iftrue), self._handle(expr.iffalse))
+        return CITE(self._handle(expr.cond), self._handle(expr.iftrue), self._handle(expr.iffalse), tags=expr.tags)
 
     def _handle_Expr_StackBaseOffset(self, expr):  # pylint:disable=no-self-use
 
         if expr.variable is not None:
-            return CUnaryOp('Reference', expr, variable=self._handle(expr.variable))
+            return CUnaryOp('Reference', expr, variable=self._handle(expr.variable), tags=expr.tags)
 
         # FIXME
-        r = CUnaryOp('Reference', expr, variable=None)
+        r = CUnaryOp('Reference', expr, variable=None, tags=expr.tags)
         r.set_type(SimTypeLongLong())
         return r
 
